/**
	Autonomous Actor:
	Autonmous Actor contains the data gathering capabilities for an individual
	Actor to learn about its surroundings (in terms of processing speed and
	message sending speed).  The Peer Theater uses this information to determine
	when and if to migrate Autonomous Actors around IO.

	Author: Travis Desell
	Last Modified: 9/20/2002
*/

package io.core.language;

import java.util.Hashtable;
import salsa.language.*;
import io.core.profiling.ActorData;
import io.core.messaging.TheaterLocator;


public class AutonomousActor extends UniversalActor implements IOActor {

	/**
	  * The following are used to mark how long an actor
	  * has been collecting data.
	  */
	private 	long lastResetTime = 0;
	private 	int messagesRecieved = 0;

	/**
	  * The following are used by the Peer Theater to determine
	  * the message sending priority and speed of an actor.
	  */
	private 	int messagesSent = 0;
	private		int foreignMessagesSent = 0;
	private 	int localMessagesSent = 0;
	private 	long messageSendingTime = 0;		//in milliseconds
	private 	Hashtable foreignReferences = new Hashtable();

	/**
	  *	The following are used by the Peer Theater to determine
	  *	the message processing priority and speed of an actor.
	  */
	private 	int messagesProcessed = 0;
	private 	long messageProcessingTime = 0;	//in milliseconds

	/**
	  *	The following is used to determine if the actor is already
	  *	scheduled to migrate away from this theater.
	  */
	protected	boolean scheduled = false;


	public AutonomousActor() {
		super();
		reset();
	}

	public void reset() {
		//after a change has been made to the state of the Autonomous
		//Actor's surroundings, reset the knowledge its gained.

		messagesRecieved = 0;
		localMessagesSent = 0;
		foreignMessagesSent = 0;
		messagesSent = 0;
		messageSendingTime = 0;
		messagesProcessed = 0;
		messageProcessingTime = 0;
		foreignReferences = new Hashtable();

		lastResetTime = System.currentTimeMillis();
	}

	public boolean scheduled() {
		return scheduled;
	}

	protected synchronized void incrementProcessed(long time) {
		messagesProcessed++;
		messageProcessingTime += time;
	}

	protected synchronized void incrementLocalSend() {
		messagesSent++;
		localMessagesSent++;
	}

	protected synchronized void incrementForeignSend(TheaterLocator targetLocation, long timeSpent) {
		messagesSent++;
		foreignMessagesSent++;

		if (foreignReferences.containsKey( targetLocation )) {
			Integer value = (Integer)foreignReferences.get(targetLocation);
			value = new Integer(value.intValue() + 1);
			foreignReferences.put(targetLocation, value);
		}
		else {
			foreignReferences.put(targetLocation, new Integer(1));
				//if we are processing messages quicker or as quick as we are
		}
	}

	public ActorData getData() {
		int localMessagesSent = messagesSent - foreignMessagesSent;
		long collectionTime = System.currentTimeMillis() - lastResetTime;

		ActorData data = new ActorData( messagesRecieved,
						messagesProcessed,
						localMessagesSent,
						foreignMessagesSent,
						messageSendingTime,
						(messageProcessingTime - messageSendingTime),
						collectionTime,
						foreignReferences );

		return data;
	}

	public boolean satisfied() {
		//we subtract one from messages recieved in case the actor is processing
		//messages as quickly as it is recieving them, and is in the middle of processing
		//a message
		
		if (actorState != null) {
			if (actorState.mailbox.isEmpty()) {
				if (messagesProcessed == 0) return true;
				else if (messagesProcessed > messagesRecieved) return true;
				else return false;
			} else {
				if (messagesProcessed > messagesRecieved) {
					return true;
				} else {
					return false;
				}
			}
		} else {
			return true;
		}
	}

	public double satisfaction() {
		if (messagesProcessed > 0) {
			return (messagesRecieved - messagesProcessed) / messagesProcessed;
		} else {
			return (messagesRecieved * 2);
		}
	}

	public void prioritySend(Message message) {
		scheduled = true;

		super.send(message);
	}

	public void send(Message message) {
		messagesRecieved++;

		Actor source = message.getSource();		

		if (actorState != null) {
			//local send
			super.send(message);
			if (source instanceof AutonomousActor)	((AutonomousActor)source).incrementLocalSend();
		} else {
			//remote send
			long time = System.currentTimeMillis();
			super.send(message);
			if (source instanceof AutonomousActor) {
				((AutonomousActor)source).incrementForeignSend(new TheaterLocator(((UniversalActor)message.getTarget()).ual.toString()), System.currentTimeMillis() - time);
			}
		}
	}
}
