package wwc.messaging; 

import java.util.Hashtable;
import java.util.PropertyResourceBundle;
import java.io.IOException;
import java.net.ServerSocket;
import java.net.Socket;
import java.net.InetAddress;
import java.net.UnknownHostException;

import salsa.language.Message;
import salsa.language.Actor;
import salsa.language.ActorState;
import salsa.language.ServiceFactory;
import salsa.language.exceptions.SalsaWrapperException;
import salsa.messaging.TheaterService;
import salsa.messaging.ReceptionService;
import salsa.naming.UAL;
import salsa.naming.UAN;


/**
 * The Theater is the reference implementation of the TheaterService
 * interface.
 *
 * @version $Id: Theater.java,v 1.6 2003/04/09 19:55:41 deselt Exp $
 *
 * @author stepha
 */
public class Theater implements TheaterService {
	private final Hashtable registry = new Hashtable();
	private final Hashtable anonymousRegistry = new Hashtable();
	
	public static String version;
	
	private ServerSocket server;
	private int port;

	private int generatedLocations;
	private PropertyResourceBundle configFile;

	/**
 	 * This method provides synchronization on the generated locations variable
 	 * to ensure that each requesting actor gets a unique location.
 	 * 
 	 * @return String
 	 */
	public synchronized UAL generateUAL() {
		UAL generated = null;

		if (this.port == 0) {
			startTheater(0, null);
		}

		try {
			generated = new UAL( "rmsp://" + InetAddress.getLocalHost().getHostName() + ":" + this.port + "/location" + (++generatedLocations));
		} catch (UnknownHostException e) {
			System.err.println("Could not generate localhost: " + e);
		}

		return generated;
	}

	/**
	 * Constructor for Theater.
	 */
	public Theater() {
		version  = "$Id: Theater.java,v 1.6 2003/04/09 19:55:41 deselt Exp $";
		server = null;
		port = 0;
		generatedLocations = 0;
	}

	public Object getEntry(UAL ual) {
		return anonymousRegistry.get(ual.toString());
	}

	public synchronized void setEntry(UAL ual, Object entry) {
		anonymousRegistry.put(ual.toString(), entry);
	}

	public Object getEntry(UAN uan) {
		return registry.get(uan.toString());
	}

	public synchronized void setEntry(UAN uan, Object entry) {
		registry.put(uan.toString(), entry);
	}

	public Hashtable getRegistry() {
		return registry;
	}

	/**
	 * @see salsa.language.TheaterService#startTheater(int, salsa.language.Actor, java.lang.String)
	 */
	public TheaterService startTheater(int port, ActorState initial) {
		if (server == null) {
			this.port = port;

			// Create a server socket.
			try {
				server = new ServerSocket( port );
			} catch (IOException e) {
				throw new SalsaWrapperException( "Could not start theater.", e );
			}
			
			// Check to see what the port actually is.
			if (port == 0) {
				this.port = server.getLocalPort();
			}
			
			final TheaterService theater = this;
			final ServerSocket passive = server;
			
			// Accept connections.
			Runnable listenLoop = new Runnable() {
				public void run() {
					int i = 0;
					while (true) {
						try {
							final Socket incoming = passive.accept();

							Runnable handlerThread = new Runnable() {
								public void run() {
									ReceptionService handler = ServiceFactory.getReception();
									handler.process(incoming);
								}
							};
							new Thread(handlerThread, "Handler Loop " + i).start();
						} catch (IOException e) {
							throw new SalsaWrapperException( "Failed to accept connection.", e );
						}
						
						i++;
					}
				}
			};
	
			// Start the listening thread.
			new Thread( listenLoop, "Listen Loop Thread" ).start();
		}

		// Add the specified actor to the theater.
		if (initial != null) setEntry( initial.getUAN(), initial );

		return this;
	}


	/**
	 * @see salsa.language.TheaterService#getPort()
	 */
	public int getPort() {
		return port;
	}


	/**
	 * used by subclassed theaters to modify the port value (from command line arguments)
	 */
	public void setPort(int _port) {
		port = _port;
	}


	public static void main( String[] args ) {
		Thread.currentThread().setName( "Main startup thread.");
		
		// Check arguments.
		if (args.length > 1) {
			System.out.println( "Theater TheaterService Reference Implementation.");
			System.out.println( "Worldwide Computing Lab. Rensselaer Polytechic Institute." );
			System.out.println( "Version: " + version );
			System.out.println( "Usage: wwc.theater.Theater [<port>]" );
			System.out.println( "or invoke a Actor with VM option -Dtheater=wwc.theater.Theater" );
			return;
		}
		
		// Determine port.
		int port = 4040;
		if (args.length == 1) {
			try {
				port = Integer.parseInt( args[0] );
			} catch (NumberFormatException e) {
				System.out.println( "Bad port." );
				System.out.println( "Theater TheaterService Reference Implementation.");
				System.out.println( "Worldwide Computing Lab. Rensselaer Polytechic Institute." );
				System.out.println( "Version: " + version );
				System.out.println( "Usage: wwc.theater.Theater [<port>]" );
				System.out.println( "or invoke a Actor with VM option -Dtheater=wwc.theater.Theater" );
				return;
			}
		}
		
		// Create a theater.
		Theater theater = new Theater();
		
		// Construct service factory.
		ServiceFactory.createFactory( theater, null, null, null );
		
		// Start the theater.
		theater.startTheater( port, null);
		
		System.out.println( "Theater started on port: " + theater.getPort());
	}
}
